# util.praat --- Praat include file containing some utilities 

# Last modification: 2014-06-30


procedure ltrim s_$
# remove all blanks at start of input string
# and return resulting string in result$
   result$ = s_$
   while (length (result$) > 0 and left$ (result$, 1) = " ")
      result$ = right$ (result$, 2, length(result$) - 1)
   endwhile
endproc


procedure rtrim s_$
# remove all blanks at end of input string
# and return resulting string in result$
   result$ = s_$
   len_ = length (result$) 
   while (len_ > 0 and right$ (result$, 1) = " ")
      result$ = left$ (result$, len_ - 1)
      len_ -= 1 
   endwhile
endproc


procedure btrim s_$
# remove all blanks at start and end of input string
# and return resulting string in result$
   result$ = s_$
   while (length (result$) > 0 and left$ (result$, 1) = " ")
      result$ = right$ (result$, 2, length(result$) - 1)
   endwhile
   len_ = length (result$) 
   while (len_ > 0 and right$ (result$, 1) = " ")
      result$ = right$ (result$, len_ - 1)
      len_ -= 1 
   endwhile
endproc


procedure files_get_regex .varname$ .spec$
# Get list of file names using a regular expression (whereas "Create Strings as file list" only allows wildcard '*').
# <spec> contains full path with possibly regex symbols in filename part.
# <varname> is the name of the variable that will hold the list object ID.
   call fname_parts '.spec$'
   .specfname$ = result1$
   .indir$ = result4$
   ; Find ALL files in the specified directory
   .listID = Create Strings as file list... filelist '.indir$'*
   .j = Get number of strings
   select .listID
   '.varname$' = .listID
   while (.j > 0)
      .s$ = Get string... .j
      .i = index_regex (.s$, .specfname$)
      if not (index_regex (.s$, .specfname$))	; Match with regular expression
         Remove string... .j
      else
         Set string... .j '.indir$''.s$'
      endif
      .j -= 1
   endwhile
   .n = Get number of strings
endproc


procedure fname_parts s_$
# Obtain filename parts
# s_$		the total filename
# result1$	the filename without path
# result2$	the basename (i.e. no path, no extension)
# result3$	the filename extension (excluding dot)
# result4$	the file path (including trailing slash) including drive
# result5$	the file path (including trailing slash) excluding drive
# result6$	the drive (excluding separator ':' )
   result1$ = s_$
   result2$ = ""
   result3$ = ""
   result4$ = ""
   result5$ = ""
   result6$ = ""
   pos_ = rindex (s_$, "\")
   if (pos_ = 0)
      pos_ = rindex (s_$, "/")
   endif
   if (pos_ > 0)
      result4$ = mid$ (s_$, 1, pos_)
      len_ = length (s_$) - pos_
      result1$ = mid$ (s_$, pos_ + 1, len_)
   endif
   pos_ = rindex (result1$, ".")
   if (pos_ > 0)
      len_ = length (result1$)
      result3$ = right$ (result1$, len_ - pos_)
      result2$ = left$ (result1$, pos_ - 1)
   else
      result2$ = result1$
   endif
   pos_ = index (result4$, ":")
   if (pos_ = 2)
      len_ = length (result4$)
      result5$ = right$ (result4$, len_ - pos_)
      result6$ = left$ (result4$, pos_ - 1)
   else
      result5$ = result4$
   endif
endproc


procedure interval_from_time gridID tiernr t varname$
# Return in <result> the interval number within tier <tiernr> of TextGrid <gridID> 
# in which time <t> occurs, taking into account cases where time is outside grid range
   select gridID
   result = Get interval at time... tiernr t
   if (result == 0) ; time outside grid range
      .tmp = Get starting time
      if (t <= .tmp)
         result = 1
      else
         .tmp = Get finishing time
         if (t >= .tmp)
            result = Get number of intervals... tiernr
         endif
      endif
   endif
   'varname$' = result
endproc


procedure tier_get_label_at_time .grid .tier .time .varname$
; for interval tier
   '.varname$'$ = ""
   select .grid
   .interval_tier = Is interval tier... .tier
   if (.interval_tier)
      .i = Get interval at time... .tier .time
      if (.i > 0)	
         '.varname$'$ = Get label of interval... .tier .i
      endif
   endif
endproc


procedure tier_get_label_in_range .grid .tier .t1 .t2 .varname$
; for point tier
;call msg tglir: grid='.grid' tier='.tier' t1='.t1:3' t2='.t2:3' varname='.varname$'
   '.varname$'$ = ""
   select .grid
   .interval_tier = Is interval tier... .tier
   if (not .interval_tier) 
      .k = Get nearest index from time... .tier .t1+(.t2-.t1)/2
      .t = Get time of point... .tier .k
;call msg tglir: k='.k' .t='.t:3'
      if (.t > .t1 and .t <=.t2)
         '.varname$'$ = Get label of point... .tier .k
      endif
   endif
endproc


procedure tier_number_by_name gridID name$
# Return in <result> the number of first tier corresponding to tier name.
# Return 0 if tier with name does not exist in textgrid.
# The target tier name may be a regular expression.
# Return tier name in <result2$>
   select gridID
   n_ = Get number of tiers
   result = 0
   tier_ = 1
   while (tier_ <= n_ and result = 0)
      result2$ = Get tier name... tier_
      if (index_regex (result2$, 'name$')) 
         result = tier_
      else
         tier_ += 1
         result2$ = ""
      endif
   endwhile
endproc


procedure tier_get .gridID tiername$ varname$ message$ fatal_
# In TextGrid object <.gridID>, find a tier with a name matching the regular expression
# in <tiername$> and return its number as the value of variable named <varname$>.
# When such a tier is not found, print the error message and exit is <fatal_> is true.
   call tier_number_by_name .gridID tiername$
   if (result < 1)
      if (fatal_)
         call fatal_error 'message$'
      else
         call error_msg 'message$'
      endif
   endif
   'varname$' = result
endproc


procedure grid_show_tiers gridID
   printline Showing tier names
   select gridID
   n_ = Get number of tiers
   tier_ = 1
   while (tier_ <= n_)
      s$ = Get tier name... tier_
      printline Tier 'tier_', name= <'s$'>
      tier_ += 1
   endwhile
endproc


procedure tier_names .grid
   select .grid
   .n = Get number of tiers
   result$ = ""
   for .j to .n
       .s$ = Get tier name... .j
       if (.j > 1)
          result$ = result$ + ", "
       endif
       result$ = result$ + "'.j':" + .s$
   endfor
endproc


procedure tier_replace .srcgrid .srctier .destgrid .desttier .dest_varname$
; Replace a tier of a grid by a tier of another grid.
; For speed, when possible, extract a tier, merge it with src TextGrid and duplicate it to appropriate position.
; However, when times don't match, copy interval by interval.
   select .destgrid
   .dest_t1 = Get start time
   .dest_t2 = Get end time
   .n = Get number of tiers
   select .srcgrid
   .src_t1 = Get start time
   .src_t2 = Get end time
   .tiername$ = Get tier name... .srctier
 printline tier_replace: src='.srcgrid' dst='.destgrid' n='.n' tiername='.tiername$'
   if (.src_t1 == .dest_t1 and .src_t2 <= .dest_t2)
      .newgrid = Extract one tier... .srctier
      if (.src_t2 < .dest_t2)
         Extend time... .dest_t2-.src_t2 End
        .newgrid_t2 = Get end time
 printline tier_replace: new endtime='.newgrid_t2' (signal_finish='signal_finish')
      endif
      select .destgrid
      plus .newgrid
      .merged = Merge
      Remove tier... .desttier
      .ntiers = Get number of tiers
      Duplicate tier... .ntiers .desttier '.tiername$'
      Remove tier... .ntiers+1
      select .destgrid
      plus .newgrid
      Remove
      '.dest_varname$' = .merged 
   else		; times don't match
      call msg tier_replace: times dont match: src '.src_t1' '.src_t2' , dest '.dest_t1' '.dest_t2'
      call copy_tier .srcgrid .srctier .destgrid .desttier
   endif
endproc


procedure tier_replace srcgrid srctier destgrid desttier
   call tier_clear destgrid desttier
   call copy_tier srcgrid srctier destgrid desttier
endproc


procedure copy_tier srcgrid srctier destgrid desttier
# assumes destination tier is empty
   select srcgrid
   interval_tier = Is interval tier... srctier
   if (interval_tier)
      select destgrid
      n = Get number of intervals... desttier
      endtime_ = Get end point... desttier n 
      select srcgrid
      n = Get number of intervals... srctier
      for i to n
         select srcgrid
         t1_ = Get start point... srctier i
         t2_ = Get end point... srctier i
         label$ = Get label of interval... srctier i
         select destgrid
         if (t2_ < endtime_ and i < n)
            Insert boundary... desttier t2_
         endif
         if (t1_ <= endtime_ )
            t2_ = min (t2_, endtime_)	; when destgrid is shorter than srcgrid
            j = Get interval at time... desttier (t1_ + (t2_ - t1_)/2)
            Set interval text... desttier j 'label$'
         endif
      endfor
   else
      select srcgrid
      n = Get number of points... srctier
      for i to n
         select srcgrid
         t = Get time of point... srctier i
         label$ = Get label of point... srctier i
         select destgrid
         Insert point... desttier t 'label$'
      endfor
   endif
endproc


procedure tier_clear gridID_ tier_
   select gridID_
   interval_tier = Is interval tier... tier_
   if (interval_tier)
      i = Get number of intervals... tier_
      i -= 1
      while (i > 0)
         Remove right boundary... tier_ i
         i -= 1
      endwhile
   else
      n_ = Get number of points... tier_
      for i to n_
         Remove point... i
      endfor
   endif
endproc


procedure tier_clear_text grid_ tier_
   select grid_
   interval_tier = Is interval tier... tier_
   if (interval_tier)
      n_ = Get number of intervals... tier_
      empty$ = ""
      for i_ from 1 to n_
         Set interval text... tier_ i_ 'empty$'
      endfor
   endif
endproc


procedure grid_append_tier gridin tierin gridname$
# append tier <tierin> from <gridin> at end of <gridname$> by creating a new grid, the ID of which will replace the value of variable <gridname$>
   select gridin
   ok = Is interval tier... tierin
   Extract tier... tierin
   if (ok)
      tmp = selected ("IntervalTier",-1)
   else
      tmp = selected ("TextTier",-1)
   endif
   grid_ = 'gridname$'
   select grid_
   plus tmp
   Append
   tmp2 = selected ("TextGrid",-1)
   select tmp
   plus grid_
   Remove
   'gridname$' = tmp2
endproc


procedure tier_match_times .grid .tier1 .tier2
   select .grid
   .n1 = Get number of intervals... .tier1
   .n2 = Get number of intervals... .tier2
   .s1$ = Get tier name... .tier1
   .s2$ = Get tier name... .tier2
   .n = min (.n1, .n2)
   for .i to .n
      .t1 = Get start point... .tier1 .i
      .t2 = Get start point... .tier2 .i
      if (.t1 <> .t2)
         printline Comparing times in tiers '.s1$' and '.s2$': first mismatch at time '.t1:3' 
         .i = .n
      endif
   endfor
endproc


procedure next_field s_$
# Get next field form a string with comma-separated fields.
# Return success in <result>, return field in <result2$>, return rest of string in <result3$>
   result2$ = ""
   len_ = length (s_$) 
   while (len_ > 0 and left$ (s_$, 1) = " ")	; ltrim input
      s_$ = right$ (s_$, len_ - 1)
      len_ -= 1 
   endwhile
   if (len_ == 0)
      result = 0
   else
      pos_ = index (s_$, ",")
      if (pos_ = 0)
         result3$ = ""
      else
         result3$ = right$ (s_$, len_ - pos_)		; rest of strings
         s_$ = left$ (s_$, pos_ - 1)
         len_ = length (s_$)
      endif
      while (len_ > 0 and left$ (s_$, 1) = " ")		; ltrim field
         s_$ = right$ (s_$, len_ - 1)
         len_ -= 1 
      endwhile
      while (len_ > 0 and right$ (s_$, 1) = " ")	; rtrim field
         s_$ = left$ (s_$, len_ - 1)
        len_ -= 1 
      endwhile
      if (length (s_$) == 0)
         result = 0
      else
         result = 1
         result2$ = s_$
      endif
   endif
endproc


procedure next_line stringname$ varname$
# Get next line form a string in variable named <stringname>
# Return line in variable named <varname$> and remainder of string in variable named <stringname$>
   .text$ = 'stringname$'$
   .len = length(.text$)
   .i = index(.text$, newline$)
   if (.i = 0)
      .i = .len
   endif
   'varname$'$ = left$(.text$, .i)
   'stringname$'$ = right$(.text$, .len-.i)
   .text$ = ""
endproc


procedure is_number s_$
    result = extractNumber(s_$, "")
    if (result == undefined)
       result = 0
    else
       result = 1
    endif
endproc


procedure convert_Hz_ST objectID
# Convert a PitchTier from Hz values to ST scale, relative to 1 Hz
   yoffset = hertzToSemitones(1)
   select objectID
   n = Get number of points
   for i from 1 to n
      x = Get time from index... i
      y = Get value at index... i
      Remove point... i
      yST =  hertzToSemitones(y) - yoffset
      Add point... x yST
   endfor
endproc


procedure convert_ST_Hz objectID
# Convert a PitchTier from ST values (relative to 1 Hz) to Hz values 
   yoffset = hertzToSemitones(1)
   select objectID
   n = Get number of points
   for i from 1 to n
      x = Get time from index... i
      y = Get value at index... i
      Remove point... i
      y =  semitonesToHertz(y + yoffset)
      Add point... x y
   endfor
endproc


procedure from_ST_rel_1_to_Hz .value
   result = semitonesToHertz(.value + hertzToSemitones(1))
endproc


procedure from_Hz_to_ST_rel_1 .value
   result = hertzToSemitones(.value) - hertzToSemitones(1) 
endproc

# ------------------------------ Messages ------------------------------ #



procedure logging_start .option$ logfile$
# Write messages (calls to msg, debug_msg, time_msg, error_msg, fatal_error) to logfile.
# Option "reset" resets the logfile (whereas otherwise new messages are appended to the logfile, if it exists)
# Option "timed" selects the output format where timing information is added to the message
# Option "noinfo" suppresses output to Info window and writes messages to logfile only.
   logging = 1
   logging_timed = 0
   logging_noinfo = 0
   if (index (.option$, "timed"))
      logging_timed = 1
   endif
   if (index (.option$, "reset"))
      filedelete 'logfile$'
   endif
   if (index (.option$, "noinfo"))
      logging_noinfo = 1
   endif
   call msg Logging to 'logfile$' started
endproc


procedure msg .message$
   if (variableExists ("logging"))
      if (logging)
         if (logging_timed)
            date_$ = date$ ()
            fileappend 'logfile$' 'tab$'@ 'date_$''newline$'
         endif
         fileappend 'logfile$' '.message$''newline$'
         if (not logging_noinfo)
            printline '.message$'
         endif
      else
         printline '.message$'
      endif
   else
      printline '.message$'
   endif
endproc


procedure debug_msg message_$
   if (variableExists ("debug"))
      if (debug)
         call msg message_$
      endif
   endif
endproc


procedure time_msg message_$
   date_$ = date$ ()
   call msg 'date_$' 'message_$'
endproc


procedure error_msg message$
   call msg *** ERROR *** 'message$'
endproc


procedure fatal_error message$
   call msg *** ERROR *** 'message$'
   exit
endproc


# ------------------------------ Miscellaneous ------------------------------ #


procedure toggle varname$
   if ('varname$' == 0)
      'varname$' = 1
   else
      'varname$' = 0
   endif
endproc


procedure play_part objectID starttime endtime
     select objectID
     tmpsoundID = Extract part... starttime endtime Rectangular 1.0 yes
     ; Rename... play_part
     Play
     Remove
endproc


